from typing import ClassVar, Iterable

from pytest_mock_resources.compat import pymongo
from pytest_mock_resources.config import DockerContainerConfig, fallback
from pytest_mock_resources.container.base import ContainerCheckFailed


class MongoConfig(DockerContainerConfig):
    """Define the configuration object for mongo.

    Args:
        image (str): The docker image:tag specifier to use for mongo containers.
            Defaults to :code:`"mongo:3.6"`.
        host (str): The hostname under which a mounted port will be available.
            Defaults to :code:`"localhost"`.
        port (int): The port to bind the container to.
            Defaults to :code:`28017`.
        ci_port (int): The port to bind the container to when a CI environment is detected.
            Defaults to :code:`27017`.
        root_database (str): The name of the root mongo database to create.
            Defaults to :code:`"dev-mongo"`.
    """

    name = "mongo"

    _fields: ClassVar[Iterable] = {"image", "host", "port", "ci_port", "root_database"}
    _fields_defaults: ClassVar[dict] = {
        "image": "mongo:3.6",
        "port": 28017,
        "ci_port": 27017,
        "root_database": "dev-mongo",
    }

    @fallback
    def root_database(self):
        raise NotImplementedError()

    def ports(self):
        return {27017: self.port}

    def check_fn(self):
        try:
            client = pymongo.MongoClient(self.host, self.port)
            db = client[self.root_database]
            db.command("ismaster")
        except pymongo.errors.ConnectionFailure:
            raise ContainerCheckFailed(
                f"Unable to connect to a presumed MongoDB test container via given config: {self}"
            )



from __future__ import annotations

import argparse
import sys

from pytest_mock_resources.config import DockerContainerConfig
from pytest_mock_resources.container.base import container_name, get_container
from pytest_mock_resources.hooks import get_docker_client
from pytest_mock_resources.plugin import find_entrypoints, load_entrypoints


class StubPytestConfig:
    pmr_multiprocess_safe = False
    pmr_cleanup_container = False
    pmr_docker_client = None

    class option:  # noqa: N801
        pmr_multiprocess_safe = False
        pmr_cleanup_container = False

    def getini(self, attr):
        return getattr(self, attr)


def main():
    entrypoints = find_entrypoints()
    load_entrypoints(entrypoints)

    parser = create_parser()
    args = parser.parse_args()

    if args.load:
        load_entrypoints(args.load)

    pytestconfig = StubPytestConfig()

    stop = args.stop
    start = not stop

    for fixture in args.fixtures:
        if fixture not in DockerContainerConfig.subclasses:
            valid_options = ", ".join(DockerContainerConfig.subclasses)
            raise argparse.ArgumentError(
                args.fixtures,
                f"'{fixture}' invalid. Valid options include: {valid_options}",
            )

        execute(fixture, pytestconfig, start=start, stop=stop)


def create_parser():
    parser = argparse.ArgumentParser(
        description="Preemptively run docker containers to speed up initial startup of PMR Fixtures."
    )
    parser.add_argument(
        "fixtures",
        metavar="Fixture",
        type=str,
        nargs="+",
        help="Available Fixtures: {}".format(", ".join(DockerContainerConfig.subclasses)),
    )
    parser.add_argument(
        "--stop", action="store_true", help="Stop previously started PMR containers"
    )

    parser.add_argument(
        "--load",
        action="append",
        help="Import a module in order to load 3rd party resources.",
    )
    return parser


def execute(fixture: str, pytestconfig: StubPytestConfig, start=True, stop=False):
    config_cls = DockerContainerConfig.subclasses[fixture]
    config = config_cls()

    if start:
        generator = get_container(pytestconfig, config)
        for _ in generator:
            pass

    if stop:
        docker = get_docker_client(config)

        assert config.port
        name = container_name(fixture, int(config.port))
        try:
            container = docker.container.inspect(name)
        except Exception:
            sys.stderr.write(f"Failed to stop {fixture} container\n")
        else:
            container.kill()


if __name__ == "__main__":
    main()



import logging
from typing import cast

import pytest
import sqlalchemy
from sqlalchemy import text
from sqlalchemy.engine import Connection, Engine

from pytest_mock_resources.container.base import (
    async_retry,
    DEFAULT_RETRIES,
    get_container,
    retry,
)
from pytest_mock_resources.container.postgres import (
    get_sqlalchemy_engine,
    PostgresConfig,
)
from pytest_mock_resources.fixture.base import asyncio_fixture, generate_fixture_id
from pytest_mock_resources.sqlalchemy import (
    bifurcate_actions,
    EngineManager,
    normalize_actions,
)

log = logging.getLogger(__name__)


class DatabaseExistsError(RuntimeError):
    """Raise when the database being created already exists.

    This exception commonly occurs during multiprocess test execution, as a
    sentinel for gracefully continuing among test workers which attempt to create
    the same database.
    """


@pytest.fixture(scope="session")
def pmr_postgres_config():
    """Override this fixture with a :class:`PostgresConfig` instance to specify different defaults.

    Examples:
        >>> @pytest.fixture(scope='session')
        ... def pmr_postgres_config():
        ...     return PostgresConfig(image="postgres:9.6.10", root_database="foo")
    """
    return PostgresConfig()


@pytest.fixture(scope="session")
def pmr_postgres_container(pytestconfig, pmr_postgres_config: PostgresConfig):
    yield from get_container(pytestconfig, pmr_postgres_config)


def create_postgres_fixture(
    *ordered_actions,
    scope="function",
    tables=None,
    session=None,
    async_=False,
    createdb_template="template1",
    engine_kwargs=None,
    template_database=True,
    actions_share_transaction=None,
):
    """Produce a Postgres fixture.

    Any number of fixture functions can be created. Under the hood they will all share the same
    database server.

    Arguments:
        ordered_actions: Any number of ordered actions to be run on test setup.
        scope: Passthrough pytest's fixture scope.
        tables: Subsets the tables created by `ordered_actions`. This is generally
            most useful when a model-base was specified in `ordered_actions`.
        session: Whether to return a session instead of an engine directly. This can
            either be a bool or a callable capable of producing a session.
        async_: Whether to return an async fixture/client.
        createdb_template: The template database used to create sub-databases. "template1" is the
            default chosen when no template is specified.
        engine_kwargs: Optional set of kwargs to send into the engine on creation.
        template_database: Defaults to True. When True, amortizes the cost of performing database
            setup through `ordered_actions`, by performing them once into a postgres "template"
            database, then creating all subsequent per-test databases from that template.
        actions_share_transaction: When True, the transaction used by `ordered_actions` context
            will be the same as the one handed to the test function. This is required in order
            to support certain usages of `ordered_actions, such as the creation of temp tables
            through a `Statements` object. By default, this behavior is enabled for synchronous
            fixtures for backwards compatibility; and disabled by default for
            asynchronous fixtures (the way v2-style/async features work in SQLAlchemy can lead
            to bad default behavior).
    """
    fixture_id = generate_fixture_id(enabled=template_database, name="pg")

    engine_kwargs_ = engine_kwargs or {}
    engine_manager_kwargs = {
        "ordered_actions": ordered_actions,
        "tables": tables,
        "createdb_template": createdb_template,
        "session": session,
        "fixture_id": fixture_id,
        "actions_share_transaction": actions_share_transaction,
    }

    @pytest.fixture(scope=scope)
    def _sync(*_, pmr_postgres_container, pmr_postgres_config):
        fixture = _sync_fixture(pmr_postgres_config, engine_manager_kwargs, engine_kwargs_)
        for _, conn in fixture:
            yield conn

    async def _async(*_, pmr_postgres_container, pmr_postgres_config):
        fixture = _async_fixture(pmr_postgres_config, engine_manager_kwargs, engine_kwargs_)
        async for _, conn in fixture:
            yield conn

    if async_:
        return asyncio_fixture(_async, scope=scope)
    return _sync


def _sync_fixture(pmr_config, engine_manager_kwargs, engine_kwargs, *, fixture="postgres"):
    root_engine = cast(Engine, get_sqlalchemy_engine(pmr_config, pmr_config.root_database))
    conn = retry(root_engine.connect, retries=DEFAULT_RETRIES)
    conn.close()
    root_engine.dispose()

    root_engine = cast(
        Engine,
        get_sqlalchemy_engine(pmr_config, pmr_config.root_database, autocommit=True),
    )
    with root_engine.connect() as root_conn:
        with root_conn.begin() as trans:
            template_database, template_manager, engine_manager = create_engine_manager(
                root_conn, **engine_manager_kwargs, fixture=fixture
            )
            trans.commit()
    root_engine.dispose()

    if template_manager:
        assert template_database

        template_engine = cast(
            Engine,
            get_sqlalchemy_engine(pmr_config, template_database, **engine_kwargs),
        )
        with template_engine.connect() as conn:
            with conn.begin() as trans:
                template_manager.run_static_actions(conn)
                trans.commit()
        template_engine.dispose()

    # Everything below is normal per-test context. We create a brand new database/engine/manager
    # distinct from what might have been used for the template database.
    root_engine = cast(
        Engine,
        get_sqlalchemy_engine(pmr_config, pmr_config.root_database, autocommit=True),
    )
    with root_engine.connect() as root_conn:
        with root_conn.begin() as trans:
            database_name = _produce_clean_database(root_conn, createdb_template=template_database)
            trans.commit()
    root_engine.dispose()

    engine = get_sqlalchemy_engine(pmr_config, database_name, **engine_kwargs)
    yield from engine_manager.manage_sync(engine)


async def _async_fixture(pmr_config, engine_manager_kwargs, engine_kwargs, *, fixture="postgres"):
    root_engine = get_sqlalchemy_engine(
        pmr_config, pmr_config.root_database, async_=True, autocommit=True
    )

    root_conn = await async_retry(root_engine.connect, retries=DEFAULT_RETRIES)
    await root_conn.close()

    async with root_engine.connect() as root_conn:
        async with root_conn.begin() as trans:
            (
                template_database,
                template_manager,
                engine_manager,
            ) = await root_conn.run_sync(
                create_engine_manager, **engine_manager_kwargs, fixture=fixture
            )
            await trans.commit()

    if template_manager:
        assert template_database

        engine = get_sqlalchemy_engine(pmr_config, template_database, **engine_kwargs, async_=True)
        async with engine.begin() as conn:
            await conn.run_sync(template_manager.run_static_actions)
            await conn.commit()
        await engine.dispose()

    # Everything below is normal per-test context. We create a brand new database/engine/manager
    # distinct from what might have been used for the template database.
    async with root_engine.connect() as root_conn:
        async with root_conn.begin() as trans:
            database_name = await root_conn.run_sync(
                _produce_clean_database, createdb_template=template_database
            )
            await trans.commit()

    await root_engine.dispose()

    engine = get_sqlalchemy_engine(pmr_config, database_name, **engine_kwargs, async_=True)
    async for engine, conn in engine_manager.manage_async(engine):
        yield engine, conn


def create_engine_manager(
    root_connection,
    *,
    ordered_actions,
    session,
    tables,
    createdb_template="template1",
    fixture_id=None,
    actions_share_transaction=None,
    fixture="postgres",
):
    normalized_actions = normalize_actions(ordered_actions, fixture=fixture)
    static_actions, dynamic_actions = bifurcate_actions(normalized_actions)

    template_database = createdb_template
    template_manager = None
    if fixture_id:
        try:
            template_database = _produce_clean_database(
                root_connection,
                createdb_template=createdb_template,
                database_name=fixture_id,
                ignore_failure=True,
            )
        except DatabaseExistsError:
            # This implies it's been created during a prior test, and the template should already be
            # populated.
            pass
        else:
            template_manager = EngineManager(
                dynamic_actions,
                static_actions=static_actions,
                session=session,
                tables=tables,
            )

        # The template to be used needs to be changed for the downstream database to the template
        # we created earlier.
        template_database = fixture_id

        # With template databases, static actions must be zeroed out so they're only executed once.
        # It only happens in this condition, so that when template databases are **not** used, we
        # execute them during the normal `manage_sync` flow per-test.
        static_actions = []

    fixture_manager = EngineManager(
        dynamic_actions,
        static_actions=static_actions,
        session=session,
        tables=tables,
        actions_share_transaction=actions_share_transaction,
    )
    return template_database, template_manager, fixture_manager


def _produce_clean_database(
    root_conn: Connection,
    *,
    trans=None,
    createdb_template="template1",
    database_name=None,
    ignore_failure=False,
):
    if not database_name:
        database_name = _generate_database_name(root_conn)

    try:
        root_conn.execute(text(f'CREATE DATABASE "{database_name}" template={createdb_template}'))
        root_conn.execute(
            text(f'GRANT ALL PRIVILEGES ON DATABASE "{database_name}" TO CURRENT_USER')
        )
    except (sqlalchemy.exc.IntegrityError, sqlalchemy.exc.ProgrammingError):
        if not ignore_failure:
            raise
        raise DatabaseExistsError()

    return database_name


def _generate_database_name(conn):
    try:
        conn.execute(text("CREATE TABLE IF NOT EXISTS pytest_mock_resource_db (id serial);"))
    except (sqlalchemy.exc.IntegrityError, sqlalchemy.exc.ProgrammingError):
        # A race condition may occur during table creation if:
        #  - another process has already created the table
        #  - the current process begins creating the table
        #  - the other process commits the table creation
        #  - the current process tries to commit the table creation
        pass

    result = conn.execute(text("INSERT INTO pytest_mock_resource_db VALUES (DEFAULT) RETURNING id"))
    id_ = next(iter(result))[0]
    return f"pytest_mock_resource_db_{id_}"



from __future__ import annotations

import contextlib
import json
import pathlib
import socket
import time
import types
from typing import Awaitable, Callable, TYPE_CHECKING, TypeVar

from pytest_mock_resources.hooks import (
    get_docker_client,
    get_pytest_flag,
    use_multiprocess_safe_mode,
)

try:
    import responses as _responses

    responses: types.ModuleType | None = _responses
    del _responses
except ImportError:
    responses = None

if TYPE_CHECKING:
    from python_on_whales.docker_client import DockerClient


DEFAULT_RETRIES = 40
DEFAULT_INTERVAL = 0.5


class ContainerCheckFailed(Exception):  # noqa: N818
    """Unable to connect to a Container."""


T = TypeVar("T")


async def async_retry(
    func: Callable[..., Awaitable[T]],
    *,
    args=(),
    kwargs={},
    retries=1,
    interval=DEFAULT_INTERVAL,
    on_exc=Exception,
) -> T:
    while retries:
        retries -= 1
        try:
            result = await func(*args, **kwargs)
        except on_exc:
            if not retries:
                raise
            time.sleep(interval)
        else:
            return result

    raise NotImplementedError()  # pragma: no cover


def retry(
    func: Callable[..., T],
    *,
    args=(),
    kwargs={},
    retries=1,
    interval=DEFAULT_INTERVAL,
    on_exc=Exception,
) -> T:
    while retries:
        retries -= 1
        try:
            result = func(*args, **kwargs)
        except on_exc:
            if not retries:
                raise
            time.sleep(interval)
        else:
            return result

    raise NotImplementedError()  # pragma: no cover


def get_container(pytestconfig, config, *, retries=DEFAULT_RETRIES, interval=DEFAULT_INTERVAL):
    multiprocess_safe_mode = use_multiprocess_safe_mode(pytestconfig)
    docker = get_docker_client(pytestconfig)

    if responses:
        # XXX: moto library may over-mock responses. SEE: https://github.com/spulec/moto/issues/1026
        responses.add_passthru("http+docker")

    # The creation of container can fail and leave us in a situation where it's
    # we will need to know whether it's been created already or not.
    container = None

    try:
        if multiprocess_safe_mode:
            from filelock import FileLock

            # get the temp directory shared by all workers (assuming pytest-xdist)
            root_tmp_dir = pytestconfig._tmp_path_factory.getbasetemp().parent
            fn = root_tmp_dir / f"pmr_create_container_{config.port}.lock"
            # wait for the container one process at a time
            with FileLock(str(fn)):
                container = wait_for_container(
                    docker,
                    config,
                    retries=retries,
                    interval=interval,
                )
            if container:
                record_container_creation(pytestconfig, container)

        else:
            container = wait_for_container(
                docker,
                config,
                retries=retries,
                interval=interval,
            )

        yield config
    finally:
        cleanup_container = get_pytest_flag(pytestconfig, "pmr_cleanup_container", default=True)
        if cleanup_container and container and not multiprocess_safe_mode:
            container.kill()


def wait_for_container(
    docker: DockerClient, config, *, retries=DEFAULT_RETRIES, interval=DEFAULT_INTERVAL
):
    """Wait for evidence that the container is up and healthy.

    The caller must provide a `check_fn` which should `raise ContainerCheckFailed` if
    it finds that the container is not yet up.
    """
    if config.port is None:
        config.set("port", unused_tcp_port())

    check_fn = config.check_fn
    run_args = (config.image,)
    run_kwargs = {
        "publish": [(dest, source) for source, dest in config.ports().items()],
        "envs": config.environment(),
        "name": container_name(config.name, config.port),
    }

    try:
        from python_on_whales.exceptions import DockerException
    except ImportError:  # pragma: no cover
        from python_on_whales.utils import DockerException

    try:
        # Perform a single attempt, for the happy-path where the container already exists.
        retry(check_fn, retries=1, interval=interval, on_exc=ContainerCheckFailed)
    except ContainerCheckFailed:
        # In the event it doesn't exist, we attempt to start the container
        try:
            container = docker.run(*run_args, **run_kwargs, detach=True, remove=True)
        except DockerException as e:
            container = None
            # This sometimes happens if multiple container fixtures race for the first
            # creation of the container, we want to still retry wait in this case.
            port_allocated_error = "port is already allocated"
            name_allocated_error = "to be able to reuse that name"

            error = str(e)
            if port_allocated_error not in error and name_allocated_error not in error:
                raise

        # And then we perform more lengthy retry cycle.
        retry(check_fn, retries=retries, interval=interval, on_exc=ContainerCheckFailed)
        return container
    return None


def container_name(name: str, port) -> str:
    return f"pmr_{name}_{port}"


def record_container_creation(pytestconfig, container):
    """Record the fact of the creation of a container.

    Record both a local reference to the container in pytest's `config` fixture,
    as well as a global PMR lock file of created containers.
    """
    pytestconfig._pmr_containers.append(container)

    fn = get_tmp_root(pytestconfig, parent=True)
    with load_container_lockfile(fn) as data:
        data.append(container.id)
        fn.write_text(json.dumps(data))


def get_tmp_root(pytestconfig, *, parent=False):
    """Get the path to the PMR lock file."""
    tmp_path_factory = pytestconfig._tmp_path_factory

    root_tmp_dir = tmp_path_factory.getbasetemp().parent
    if parent:
        root_tmp_dir = root_tmp_dir.parent

    return root_tmp_dir / "pmr.json"


@contextlib.contextmanager
def load_container_lockfile(path: pathlib.Path):
    """Produce the contents of the given file behind a file lock."""
    import filelock

    with filelock.FileLock(str(path) + ".lock"):
        if path.is_file():
            with open(path, "rb") as f:
                yield json.load(f)
        else:
            yield []


def unused_tcp_port():
    """Find an unused localhost TCP port from 1024-65535 and return it."""
    with contextlib.closing(socket.socket()) as sock:
        sock.bind(("127.0.0.1", 0))
        return sock.getsockname()[1]



from pytest_mock_resources.container.base import get_container
from pytest_mock_resources.container.mongo import MongoConfig
from pytest_mock_resources.container.moto import MotoConfig
from pytest_mock_resources.container.mysql import MysqlConfig
from pytest_mock_resources.container.postgres import PostgresConfig
from pytest_mock_resources.container.redis import RedisConfig
from pytest_mock_resources.container.redshift import RedshiftConfig

__all__ = [
    "get_container",
    "MongoConfig",
    "MysqlConfig",
    "PostgresConfig",
    "PostgresConfig",
    "RedisConfig",
    "RedshiftConfig",
    "MotoConfig",
]



import pytest

from pytest_mock_resources.compat import pymongo
from pytest_mock_resources.container.base import get_container
from pytest_mock_resources.container.mongo import MongoConfig
from pytest_mock_resources.credentials import Credentials


@pytest.fixture(scope="session")
def pmr_mongo_config():
    """Override this fixture with a :class:`MongoConfig` instance to specify different defaults.

    Examples:
        >>> @pytest.fixture(scope='session')
        ... def pmr_mongo_config():
        ...     return MongoConfig(image="mongo:3.4", root_database="foo")
    """
    return MongoConfig()


@pytest.fixture(scope="session")
def pmr_mongo_container(pytestconfig, pmr_mongo_config):
    yield from get_container(pytestconfig, pmr_mongo_config)


def create_mongo_fixture(scope="function"):
    """Produce a mongo fixture.

    Any number of fixture functions can be created. Under the hood they will all share the same
    database server.

    Arguments:
        scope: Passthrough pytest's fixture scope.
    """

    @pytest.fixture(scope=scope)
    def _(pmr_mongo_container, pmr_mongo_config):
        return _create_clean_database(pmr_mongo_config)

    return _


def _create_clean_database(config):
    root_client = pymongo.MongoClient(config.host, config.port)
    root_db = root_client[config.root_database]

    # Create a collection called `pytestMockResourceDbs' in the admin tab if not already created.
    db_collection = root_db["pytestMockResourcesDbs"]

    # Create a Document in the `pytestMockResourcesDbs` collection.
    result = db_collection.insert_one({})

    #  Create a database where the name is equal to that ID.
    db_id = str(result.inserted_id)
    new_database = root_client[db_id]

    #  Create a user as that databases owner
    password = "password"  # noqa: S105
    new_database.command("createUser", db_id, pwd=password, roles=["dbOwner"])

    #  pass back an authenticated db connection
    limited_client = pymongo.MongoClient(
        config.host, config.port, username=db_id, password=password, authSource=db_id
    )
    limited_db = limited_client[db_id]

    Credentials.assign_from_credentials(
        limited_db,
        drivername="mongodb",
        host=config.host,
        port=config.port,
        username=db_id,
        password=password,
        database=db_id,
    )
    return limited_db



import os
import warnings

_resource_kinds = ["postgres", "redshift", "mongo", "redis", "mysql", "moto"]


def pytest_addoption(parser):
    parser.addini(
        "pmr_multiprocess_safe",
        "Enables multiprocess-safe mode",
        type="bool",
        default=False,
    )
    parser.addini(
        "pmr_cleanup_container",
        "Optionally disable attempts to cleanup created containers",
        type="bool",
        default=True,
    )
    parser.addini(
        "pmr_docker_client",
        "Optional docker client name to use: docker, podman, nerdctl",
        type="string",
        default=None,
    )

    group = parser.getgroup("collect")
    group.addoption(
        "--pmr-multiprocess-safe",
        action="store_true",
        default=False,
        help="Enable multiprocess-safe mode",
        dest="pmr_multiprocess_safe",
    )
    group.addoption(
        "--pmr-cleanup-container",
        action="store_true",
        default=True,
        help="Optionally disable attempts to cleanup created containers",
        dest="pmr_cleanup_container",
    )
    group.addoption(
        "--pmr-docker-client",
        default=None,
        help="Optional docker client name to use: docker, podman, nerdctl",
        dest="pmr_docker_client",
    )


def get_pytest_flag(config, name, *, default=None):
    value = getattr(config.option, name, default)
    if value:
        return value

    return config.getini(name)


def use_multiprocess_safe_mode(config):
    return bool(get_pytest_flag(config, "pmr_multiprocess_safe"))


def get_docker_client_name(config) -> str:
    pmr_docker_client = os.getenv("PMR_DOCKER_CLIENT")
    if pmr_docker_client:
        return pmr_docker_client

    docker_client = get_pytest_flag(config, "pmr_docker_client")
    if docker_client:
        return docker_client

    import shutil

    for client_name in ["docker", "podman", "nerdctl"]:
        if shutil.which(client_name):
            break
    else:
        client_name = "docker"

    config.option.pmr_docker_client = client_name
    return client_name


def get_docker_client(config):
    from python_on_whales.docker_client import DockerClient

    client_name = get_docker_client_name(config)
    return DockerClient(client_call=[client_name])


def pytest_configure(config):
    """Register markers for each resource kind."""
    for resource_kind in _resource_kinds:
        config.addinivalue_line(
            "markers",
            f"{resource_kind}: Tests which make use of {resource_kind} fixtures",
        )

    config._pmr_containers = []


def pytest_itemcollected(item):
    """Attach markers to each test which uses a fixture of one of the resources."""
    if not hasattr(item, "fixturenames"):
        return

    fixturenames = set(item.fixturenames)
    for resource_kind in _resource_kinds:
        resource_fixture = f"pmr_{resource_kind}_container"
        if resource_fixture in fixturenames:
            item.add_marker(resource_kind)


def pytest_sessionfinish(session, exitstatus):
    config = session.config

    if not use_multiprocess_safe_mode(config):
        return

    # In the context of multiprocess pytest invocations like pytest-xdist,
    # `workerinput` will be `None` in only the root pytest call.
    workerinput = getattr(session.config, "workerinput", None)
    if workerinput is not None:
        return

    # docker-based fixtures should be optional based on the selected extras.
    try:
        docker = get_docker_client(config)
    except ImportError:
        return

    # We ought to avoid performing deep imports here, this file is auto-loaded
    # by pytest plugin machinery.
    from pytest_mock_resources.container.base import get_tmp_root, load_container_lockfile

    # Kind of a neat side-effect of using the below lock file is that if past
    # PMR runs failed to clean up their container, subsequent runs. Ironically
    # this might also lead to literal concurrent runs of unrelated PMR-enabled
    # pytest runs to clobber one another...:shrug:.
    roots = [get_tmp_root(session.config), get_tmp_root(session.config, parent=True)]
    for fn in roots:
        with load_container_lockfile(fn) as containers:
            if not containers:
                continue

            while containers:
                container_id = containers.pop(0)

                try:
                    container = docker.container.inspect(container_id)
                except Exception:
                    warnings.warn(
                        f"Unrecognized container {container_id}. You may need to manually delete/edit {fn}"
                    )
                else:
                    try:
                        container.kill()
                    except Exception:
                        warnings.warn(f"Failed to kill container {container_id}")

        fn.unlink()



import pytest
import sqlalchemy
from sqlalchemy import text

from pytest_mock_resources.container.base import get_container
from pytest_mock_resources.container.mysql import get_sqlalchemy_engine, MysqlConfig
from pytest_mock_resources.sqlalchemy import EngineManager


@pytest.fixture(scope="session")
def pmr_mysql_config():
    """Override this fixture with a :class:`MysqlConfig` instance to specify different defaults.

    Examples:
        >>> @pytest.fixture(scope='session')
        ... def pmr_mysql_config():
        ...     return MysqlConfig(image="mysql:5.2", root_database="foo")
    """
    return MysqlConfig()


@pytest.fixture(scope="session")
def pmr_mysql_container(pytestconfig, pmr_mysql_config):
    yield from get_container(pytestconfig, pmr_mysql_config, interval=1, retries=60)


def create_mysql_fixture(
    *ordered_actions,
    scope="function",
    tables=None,
    session=None,
    engine_kwargs=None,
):
    """Produce a MySQL fixture.

    Any number of fixture functions can be created. Under the hood they will all share the same
    database server.

    Arguments:
        ordered_actions: Any number of ordered actions to be run on test setup.
        scope: Passthrough pytest's fixture scope.
        tables: Subsets the tables created by `ordered_actions`. This is generally
            most useful when a model-base was specified in `ordered_actions`.
        session: Whether to return a session instead of an engine directly. This can
            either be a bool or a callable capable of producing a session.
        engine_kwargs: Optional set of kwargs to send into the engine on creation.
    """

    @pytest.fixture(scope=scope)
    def _(pmr_mysql_container, pmr_mysql_config):
        database_name = _create_clean_database(pmr_mysql_config)
        engine = get_sqlalchemy_engine(
            pmr_mysql_config,
            database_name,
            **(engine_kwargs or {}),
        )

        engine_manager = EngineManager.create(
            fixture="mysql",
            dynamic_actions=ordered_actions,
            tables=tables,
            session=session,
        )
        for _, conn in engine_manager.manage_sync(engine):
            yield conn

    return _


def _create_clean_database(config):
    root_engine = get_sqlalchemy_engine(config, config.root_database, isolation_level="AUTOCOMMIT")

    with root_engine.begin() as conn:
        try:
            conn.execute(
                text(
                    """
                CREATE TABLE IF NOT EXISTS pytest_mock_resource_db(
                    id serial
                );
                """
                )
            )
        except (sqlalchemy.exc.IntegrityError, sqlalchemy.exc.ProgrammingError):
            # A race condition may occur during table creation if:
            #  - another process has already created the table
            #  - the current process begins creating the table
            #  - the other process commits the table creation
            #  - the current process tries to commit the table creation
            pass

        conn.execute(text("INSERT INTO pytest_mock_resource_db VALUES (DEFAULT)"))
        result = conn.execute(text("SELECT LAST_INSERT_ID()"))
        id_ = next(iter(result))[0]
        database_name = f"pytest_mock_resource_db_{id_}"

        conn.execute(text(f"CREATE DATABASE {database_name}"))

    return database_name



import contextlib
from unittest import mock

from sqlalchemy.sql.base import Executable

from pytest_mock_resources.container.postgres import PostgresConfig
from pytest_mock_resources.patch.redshift.mock_s3_copy import mock_s3_copy_command, strip
from pytest_mock_resources.patch.redshift.mock_s3_unload import mock_s3_unload_command


@contextlib.contextmanager
def patch_connect(config: PostgresConfig, database: str):
    try:
        # Not all consumers of redshift may be using psycopg2, so it could be unavailable.
        import psycopg2
    except ImportError:
        yield
        return

    new_connect = mock_psycopg2_connect(config, database, _connect=psycopg2._connect)

    # We patch `psycopg2._connect` specifically because it allows us to patch the
    # connection regardless of the import style used by the caller.
    with mock.patch("psycopg2._connect", new=new_connect) as p:
        yield p


def mock_psycopg2_connect(config: PostgresConfig, database: str, _connect):
    """Patch `psycopg2._connect`.

    Add support for S3 COPY and UNLOAD.
    """
    import psycopg2

    class CustomCursor(psycopg2.extensions.cursor):
        """A custom cursor class to define a custom execute method."""

        def execute(self, sql, args=None):
            if isinstance(sql, Executable):
                return super().execute(sql, args)

            if strip(sql).lower().startswith("copy"):
                mock_s3_copy_command(sql, self)
                sql = "commit"

            if strip(sql).lower().startswith("unload"):
                mock_s3_unload_command(sql, self)
                sql = "commit"

            return super().execute(sql, args)

    def _mock_psycopg2_connect(*args, **kwargs):
        """Substitute the default cursor with a custom cursor."""
        conn = _connect(*args, **kwargs)
        dsn_info = conn.get_dsn_parameters()

        # We want to be sure to *only* patch the cursor's behavior when we think the
        # database connection is for the database we're specifically referencing. This
        # should prevent over-patching for connections which are not relevant to our
        # fixture.
        connection_info_matches = (
            config.host == dsn_info["host"]
            and str(config.port) == dsn_info["port"]
            and database == dsn_info["dbname"]
        )

        if connection_info_matches:
            conn.cursor_factory = CustomCursor
        return conn

    return _mock_psycopg2_connect



from __future__ import annotations

import abc
from typing import ClassVar, Iterable


class AbstractAction(metaclass=abc.ABCMeta):
    fixtures: ClassVar[tuple[str, ...]] = ()
    static_safe: ClassVar[bool] = False

    @abc.abstractmethod
    def apply(self, conn):
        """Execute an action against the provided fixture connection."""


def validate_actions(actions, *, fixture: str | None, additional_types: Iterable = ()):
    for action in actions:
        if not isinstance(action, (AbstractAction, *additional_types)):
            extra_types_str = ", ".join(
                ["`" + ".".join([cls.__module__, cls.__name__]) + "`" for cls in additional_types]
            )
            raise ValueError(
                f"`{action}` invalid: create_{fixture}_fixture function accepts "
                f"{extra_types_str}, or `AbstractAction` subclasses as inputs."
            )

        if fixture and isinstance(action, AbstractAction) and fixture not in action.fixtures:
            supported_fixtures = ", ".join([f"`create_{f}_fixture`" for f in action.fixtures])
            raise ValueError(
                f"`{action}` invalid: `{action.__class__}` is being used with `create_{fixture}_fixture` "
                f"but only supports: {supported_fixtures}."
            )



from typing import ClassVar, Iterable

import sqlalchemy

from pytest_mock_resources import compat
from pytest_mock_resources.config import DockerContainerConfig, fallback
from pytest_mock_resources.container.base import ContainerCheckFailed


class MysqlConfig(DockerContainerConfig):
    """Define the configuration object for MySql.

    Args:
        image (str): The docker image:tag specifier to use for mysql containers.
            Defaults to :code:`"mysql:5.6"`.
        host (str): The hostname under which a mounted port will be available.
            Defaults to :code:`"localhost"`.
        port (int): The port to bind the container to.
            Defaults to :code:`5532`.
        ci_port (int): The port to bind the container to when a CI environment is detected.
            Defaults to :code:`5432`.
        username (str): The username of the root user
            Defaults to :code:`"user"`.
        password (str): The password of the root password
            Defaults to :code:`"password"`.
        root_database (str): The name of the root database to create.
            Defaults to :code:`"dev"`.
    """

    name = "mysql"
    _fields: ClassVar[Iterable] = {
        "image",
        "host",
        "port",
        "ci_port",
        "username",
        "password",
        "root_database",
    }
    _fields_defaults: ClassVar[dict] = {
        "image": "mysql:5.6",
        "port": 3406,
        "ci_port": 3306,
        # XXX: For now, username is disabled/ignored. We need root access for PMR
        #      internals.
        "username": "root",
        "password": "password",
        "root_database": "dev",
    }

    @fallback
    def username(self):
        raise NotImplementedError()

    @fallback
    def password(self):
        raise NotImplementedError()

    @fallback
    def root_database(self):
        raise NotImplementedError()

    def ports(self):
        return {3306: self.port}

    def environment(self):
        return {
            "MYSQL_DATABASE": self.root_database,
            "MYSQL_ROOT_PASSWORD": self.password,
        }

    def check_fn(self):
        try:
            get_sqlalchemy_engine(self, self.root_database)
        except sqlalchemy.exc.OperationalError:
            raise ContainerCheckFailed(
                f"Unable to connect to a presumed MySQL test container via given config: {self}"
            )


def get_sqlalchemy_engine(config, database_name, **engine_kwargs):
    url = compat.sqlalchemy.URL(
        "mysql+pymysql",
        username=config.username,
        password=config.password,
        host=config.host,
        port=config.port,
        database=database_name,
    )

    from pytest_mock_resources.compat import pymysql

    pymysql.connect

    engine = sqlalchemy.create_engine(url, **engine_kwargs)

    # Verify engine is connected
    engine.connect()

    return engine



from typing import Optional

import sqlalchemy
import sqlalchemy.engine.url
from sqlalchemy.schema import MetaData

from pytest_mock_resources.compat.import_ import ImportAdaptor

version = getattr(sqlalchemy, "__version__", "")


if version.startswith("1.4") or version.startswith("2."):
    from sqlalchemy.ext import asyncio
    from sqlalchemy.orm import declarative_base, DeclarativeMeta

    URL = sqlalchemy.engine.url.URL.create

    select = sqlalchemy.select
else:
    from sqlalchemy.ext.declarative import declarative_base, DeclarativeMeta

    URL = sqlalchemy.engine.url.URL  # type: ignore[assignment]

    asyncio = ImportAdaptor(
        "SQLAlchemy",
        "SQLAlchemy >= 1.4",
        fail_message="Cannot use sqlalchemy async features with SQLAlchemy < 1.4.\n",
    )

    def _select(*args, **kwargs):
        return sqlalchemy.select(list(args), **kwargs)

    select = _select


def extract_model_base_metadata(base) -> Optional[sqlalchemy.MetaData]:
    metadata = getattr(base, "metadata", None)
    if isinstance(metadata, MetaData):
        return metadata

    return None


__all__ = [
    "asyncio",
    "declarative_base",
    "DeclarativeMeta",
    "URL",
    "select",
    "version",
]



from sqlalchemy.orm import Session

from pytest_mock_resources import compat


class Credentials:
    """Return as `pmr_credentials` attribute on supported docker-based fixtures.

    Examples:
        It's also directly dict-able.
        >>> creds = Credentials('d', 'l', 'p', 'baz', 'user', 'pass')
        >>> dict_creds = dict(creds)
    """

    def __init__(self, drivername, host, port, database, username, password):
        self.drivername = drivername
        self.host = host
        self.port = port
        self.database = str(database)
        self.username = username
        self.password = password

    def __iter__(self):
        for item in self.__dict__:
            yield (item, self[item])

    def __getitem__(self, item):
        return self.__dict__[item]

    def as_url(self):
        """Return a stringified dbapi URL string."""
        return self.as_sqlalchemy_url().render_as_string(hide_password=False)

    def as_sqlalchemy_url(self):
        """Return a sqlalchemy :class:`sqlalchemy.engine.url.URL`."""
        return compat.sqlalchemy.URL(
            drivername=self.drivername,
            host=self.host,
            port=self.port,
            database=self.database,
            username=self.username,
            password=self.password,
        )

    def as_sqlalchemy_url_kwargs(self):
        """Return the valid arguments to sqlalchemy :class:`sqlalchemy.engine.url.URL`."""
        return dict(self)

    def as_psycopg2_kwargs(self):
        """Return the valid arguments to sqlalchemy :class:`psycopg2.connect`."""
        return {
            "host": self.host,
            "port": self.port,
            "user": self.username,
            "password": self.password,
            "dbname": self.database,
        }

    def as_mongo_kwargs(self):
        """Return the valid arguments to a mongo client."""
        return {
            "host": self.host,
            "port": self.port,
            "username": self.username,
            "password": self.password,
            "authSource": self.database,
        }

    def as_redis_kwargs(self):
        """Return the valid arguments to a redis client."""
        return {
            "host": self.host,
            "port": self.port,
            "db": int(self.database),
            "username": self.username,
            "password": self.password,
        }

    @classmethod
    def assign_from_connection(cls, connection):
        if isinstance(connection, Session):
            url = connection.connection().engine.url
        else:
            url = connection.url

        instance = cls(
            drivername=url.drivername,
            host=url.host,
            port=url.port,
            username=url.username,
            password=url.password,
            database=url.database,
        )
        connection.pmr_credentials = instance
        return instance

    @classmethod
    def assign_from_credentials(cls, engine, **credentials):
        instance = Credentials(**credentials)
        engine.pmr_credentials = instance
        return instance



import sys
from typing import ClassVar, Iterable, Optional

import sqlalchemy
import sqlalchemy.exc

from pytest_mock_resources.compat.sqlalchemy import URL
from pytest_mock_resources.config import DockerContainerConfig, fallback
from pytest_mock_resources.container.base import ContainerCheckFailed

if sys.version_info < (3, 8):
    from importlib_metadata import Distribution
else:
    from importlib.metadata import Distribution


class PostgresConfig(DockerContainerConfig):
    """Define the configuration object for postgres.

    Args:
        image (str): The docker image:tag specifier to use for postgres containers.
            Defaults to :code:`"postgres:9.6.10-alpine"`.
        host (str): The hostname under which a mounted port will be available.
            Defaults to :code:`"localhost"`.
        port (int): The port to bind the container to.
            Defaults to :code:`5532`.
        ci_port (int): The port to bind the container to when a CI environment is detected.
            Defaults to :code:`5432`.
        username (str): The username of the root postgres user
            Defaults to :code:`"user"`.
        password (str): The password of the root postgres password
            Defaults to :code:`"password"`.
        root_database (str): The name of the root postgres database to create.
            Defaults to :code:`"dev"`.
        drivername (str): The sqlalchemy driver to use
            Defaults to :code:`"postgresql+psycopg2"`.
    """

    name = "postgres"
    _fields: ClassVar[Iterable] = {
        "image",
        "host",
        "port",
        "ci_port",
        "username",
        "password",
        "root_database",
        "drivername",
    }
    _fields_defaults: ClassVar[dict] = {
        "image": "postgres:9.6.10-alpine",
        "port": 5532,
        "ci_port": 5432,
        "username": "user",
        "password": "password",
        "root_database": "dev",
        "drivername": None,
    }

    @fallback
    def username(self):
        raise NotImplementedError()

    @fallback
    def password(self):
        raise NotImplementedError()

    @fallback
    def root_database(self):
        raise NotImplementedError()

    @fallback
    def drivername(self):
        raise NotImplementedError()

    def ports(self):
        return {5432: self.port}

    def environment(self):
        return {
            "POSTGRES_DB": self.root_database,
            "POSTGRES_USER": self.username,
            "POSTGRES_PASSWORD": self.password,
        }

    def check_fn(self):
        import socket

        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        try:
            s.connect((self.host, int(self.port)))
        except (OSError, ConnectionRefusedError):
            raise ContainerCheckFailed(
                f"Unable to connect to a presumed Postgres test container via given config: {self}"
            )
        finally:
            s.close()


def get_sqlalchemy_engine(config, database_name, async_=False, autocommit=False, **engine_kwargs):
    # For backwards compatibility, our hardcoded default is psycopg2, and async fixtures
    # will not work with psycopg2, so we instead swap the default to the preferred async driver.
    drivername = detect_driver(config.drivername, async_=async_)

    url = URL(
        drivername=drivername,
        host=config.host,
        port=config.port,
        username=config.username,
        password=config.password,
        database=database_name,
    )

    if autocommit:
        engine_kwargs["isolation_level"] = "AUTOCOMMIT"

    if getattr(url.get_dialect(), "is_async", None):
        from sqlalchemy.ext.asyncio import create_async_engine

        engine = create_async_engine(url, **engine_kwargs)
    else:
        engine = sqlalchemy.create_engine(url, **engine_kwargs)

    return engine


def detect_driver(drivername: Optional[str] = None, async_: bool = False) -> str:
    if drivername:
        return drivername

    if any(Distribution.discover(name="psycopg")):
        return "postgresql+psycopg"

    if async_:
        if any(Distribution.discover(name="asyncpg")):
            return "postgresql+asyncpg"
    else:
        if any(Distribution.discover(name="psycopg2")) or any(
            Distribution.discover(name="psycopg2-binary")
        ):
            return "postgresql+psycopg2"

    raise ValueError(  # pragma: no cover
        "No suitable driver found for Postgres. Please install `psycopg`, `psycopg2`, "
        "`asyncpg`, or explicitly configure the `drivername=` field of `PostgresConfig`."
    )



import uuid
from typing import Union

import pytest
from typing_extensions import Literal


def generate_fixture_id(enabled: bool = True, name=""):
    if enabled:
        uuid_str = str(uuid.uuid4()).replace("-", "_")
        return "_".join(["pmr_template", name, uuid_str])
    return None


def asyncio_fixture(async_fixture, scope="function"):
    # pytest-asyncio in versions >=0.17 force you to use a `pytest_asyncio.fixture`
    # call instead of `pytest.fixture`. Given that this would introduce an unnecessary
    # dependency on pytest-asyncio (when there are other alternatives) seems less than
    # ideal, so instead we can just set the flag that they set, as the indicator.
    async_fixture._force_asyncio_fixture = True

    fixture = pytest.fixture(scope=scope)
    return fixture(async_fixture)


Scope = Union[
    Literal["session"],
    Literal["package"],
    Literal["module"],
    Literal["class"],
    Literal["function"],
]



from pytest_mock_resources.fixture.moto.action import MotoAction, S3Bucket, S3Object
from pytest_mock_resources.fixture.moto.base import (
    create_moto_fixture,
    Credentials,
    pmr_moto_config,
    pmr_moto_container,
    Session,
)

__all__ = [
    "Credentials",
    "MotoAction",
    "S3Bucket",
    "S3Object",
    "Session",
    "create_moto_fixture",
    "pmr_moto_config",
    "pmr_moto_container",
]



from __future__ import annotations

import time
from dataclasses import dataclass

import pytest

from pytest_mock_resources.action import validate_actions
from pytest_mock_resources.compat import boto3
from pytest_mock_resources.container.base import get_container
from pytest_mock_resources.container.moto import endpoint_url, MotoConfig
from pytest_mock_resources.fixture.base import Scope
from pytest_mock_resources.fixture.moto.action import apply_ordered_actions, MotoAction


@pytest.fixture(scope="session")
def pmr_moto_config():
    """Override this fixture with a :class:`MotoConfig` instance to specify different defaults.

    Examples:
        >>> @pytest.fixture(scope='session')
        ... def pmr_moto_config():
        ...     return MotoConfig(image="motoserver/moto:latest")
    """
    return MotoConfig()


@pytest.fixture(scope="session")
def pmr_moto_container(pytestconfig, pmr_moto_config):
    yield from get_container(pytestconfig, pmr_moto_config)


def create_moto_fixture(
    *ordered_actions: MotoAction,
    region_name: str = "us-east-1",
    scope: Scope = "function",
):
    """Produce a Moto fixture.

    Any number of fixture functions can be created. Under the hood they will all share the same
    moto server.

    .. note::

       Each test executes using a different (fake) AWS account through moto. If you create
       boto3 ``client``/``resource`` objects outside of the one handed to the test (for example,
       in the code under test), they should be sure to use the ``aws_access_key_id``,
       ``aws_secret_access_key``, ``aws_session_token``, and ``endpoint_url`` given by the
       ``<fixturename>.pmr_credentials`` attribute.

    .. note::

       A moto dashboard should be available for debugging while the container is running.
       By default it would be available at ``http://localhost:5555/moto-api/#`` (but
       the exact URL may be different depending on your host/port config.

    Args:
        ordered_actions: Any number of ordered actions to be run on test setup.
        region_name (str): The name of the AWS region to use, defaults to "us-east-1".
        scope (str): The scope of the fixture can be specified by the user, defaults to "function".
    """
    validate_actions(ordered_actions, fixture="moto")

    @pytest.fixture(scope=scope)
    def _fixture(pmr_moto_container, pmr_moto_config) -> Session:
        url = endpoint_url(pmr_moto_config)
        credentials = Credentials.from_endpoint_url(url, region_name=region_name)

        session = Session(
            boto3.Session(
                aws_access_key_id=credentials.aws_access_key_id,
                aws_secret_access_key=credentials.aws_secret_access_key,
                aws_session_token=credentials.aws_session_token,
                region_name=region_name,
            ),
            endpoint_url=credentials.endpoint_url,
            pmr_credentials=credentials,
        )
        apply_ordered_actions(session, ordered_actions)
        return session

    return _fixture


@dataclass
class Credentials:
    aws_access_key_id: str
    aws_secret_access_key: str
    aws_session_token: str
    endpoint_url: str
    region_name: str = "us-east-1"

    @classmethod
    def from_endpoint_url(
        cls, url: str, account_id: str | None = None, region_name: str = "us-east-1"
    ):
        if account_id is None:
            # Attempt at a cross-process way of generating unique 12-character integers.
            account_id = str(time.time_ns())[:12]

        sts = boto3.client(
            "sts",
            endpoint_url=url,
            aws_access_key_id="test",
            aws_secret_access_key="test",  # noqa: S106
            region_name=region_name,
        )
        response = sts.assume_role(
            RoleArn=f"arn:aws:iam::{account_id}:role/my-role",
            RoleSessionName="test-session-name",
            ExternalId="test-external-id",
        )

        return cls(
            aws_access_key_id=response["Credentials"]["AccessKeyId"],
            aws_secret_access_key=response["Credentials"]["SecretAccessKey"],
            aws_session_token=response["Credentials"]["SessionToken"],
            endpoint_url=url,
            region_name=region_name,
        )

    def as_kwargs(self):
        return {
            "aws_access_key_id": self.aws_access_key_id,
            "aws_secret_access_key": self.aws_secret_access_key,
            "aws_session_token": self.aws_session_token,
            "endpoint_url": self.endpoint_url,
            "region_name": self.region_name,
        }


@dataclass
class Session:
    """Wrap the vanilla boto3 Session object, automatically inserting the endpoint_url field."""

    session: boto3.Session
    endpoint_url: str
    pmr_credentials: Credentials

    def client(self, service_name, **kwargs):
        return self.session.client(service_name, endpoint_url=self.endpoint_url, **kwargs)

    def resource(self, service_name, **kwargs):
        return self.session.resource(service_name, endpoint_url=self.endpoint_url, **kwargs)



import importlib
import sys
from typing import Iterable

if sys.version_info < (3, 8):
    import importlib_metadata
else:
    import importlib.metadata as importlib_metadata


def find_entrypoints() -> Iterable[str]:
    modules = set()
    for dist in importlib_metadata.distributions():
        for ep in dist.entry_points:
            if ep.group.lower() != "pmr":
                continue

            modules.add(ep.value)
    return sorted(modules)


def load_entrypoints(modules: Iterable[str]):
    for module in modules:
        importlib.import_module(module)



from __future__ import annotations

import functools
import os
import socket
from typing import ClassVar, Iterable

_DOCKER_HOST = "host.docker.internal"


def is_ci():
    return os.getenv("CI") == "true"


@functools.lru_cache()
def is_docker_host():
    try:
        socket.gethostbyname(_DOCKER_HOST)
    except socket.gaierror:
        return False
    else:
        return True


def get_env_config(name, kind, default=None):
    env_var = f"PMR_{name.upper()}_{kind.upper()}"
    return os.environ.get(env_var, default)


def fallback(fn):
    attr = fn.__name__

    @property
    @functools.wraps(fn)
    def wrapper(self):
        value = get_env_config(self.name, attr)
        if value is not None:
            return value

        if self.has(attr):
            return self.get(attr)

        try:
            return fn(self)
        except NotImplementedError:
            if attr in self._fields_defaults:
                return self._fields_defaults[attr]
            return None

    return wrapper


class DockerContainerConfig:
    name: ClassVar[str]

    _fields: ClassVar[Iterable] = {"image", "host", "port", "ci_port", "container_args"}
    _fields_defaults: ClassVar[dict] = {}

    subclasses: ClassVar[dict[str, type[DockerContainerConfig]]] = {}

    @classmethod
    def __init_subclass__(cls):
        DockerContainerConfig.subclasses[cls.name] = cls

    def __init__(self, **kwargs):
        for field, value in kwargs.items():
            if field not in self._fields:
                continue

            attr = f"_{field}"
            setattr(self, attr, value)

    def __repr__(self):
        cls_name = self.__class__.__name__
        return "{cls_name}({attrs})".format(
            cls_name=cls_name,
            attrs=", ".join(f"{attr}={getattr(self, attr)!r}" for attr in self._fields),
        )

    def has(self, attr):
        attr_name = f"_{attr}"
        return hasattr(self, attr_name)

    def get(self, attr):
        attr_name = f"_{attr}"
        return getattr(self, attr_name)

    def set(self, attr, value):
        attr_name = f"_{attr}"
        return setattr(self, attr_name, value)

    @fallback
    def image(self):
        raise NotImplementedError()

    @fallback
    def ci_port(self):
        raise NotImplementedError()

    @fallback
    def host(self):
        if os.environ.get("PYTEST_MOCK_RESOURCES_HOST") is not None:
            return os.environ.get("PYTEST_MOCK_RESOURCES_HOST")

        if is_docker_host():
            return _DOCKER_HOST

        return "localhost"

    @fallback
    def port(self):
        ci_port = self.ci_port
        if ci_port and is_ci():
            return ci_port

        raise NotImplementedError()

    @fallback
    def container_args(self):
        return ()

    def ports(self):
        return {}

    def environment(self):
        return {}

    def check_fn(self):
        pass



import pytest
from pytest_mock_resources.compat import redis
from pytest_mock_resources.container.base import get_container
from pytest_mock_resources.container.redis import RedisConfig
from pytest_mock_resources.credentials import Credentials


@pytest.fixture(scope="session")
def pmr_redis_config():
    """Override this fixture with a :class:`RedisConfig` instance to specify different defaults.

    Examples:
        >>> @pytest.fixture(scope='session')
        ... def pmr_redis_config():
        ...     return RedisConfig(image="redis:6.0")
    """
    return RedisConfig()


@pytest.fixture(scope="session")
def pmr_redis_container(pytestconfig, pmr_redis_config):
    yield from get_container(pytestconfig, pmr_redis_config)


def create_redis_fixture(scope="function", decode_responses: bool = False):
    """Produce a Redis fixture.

    Any number of fixture functions can be created. Under the hood they will all share the same
    database server.

    .. note::

       If running tests in parallel, the implementation fans out to different redis "database"s,
       up to a 16 (which is the default container fixed limit). This means you can only run
       up to 16 simultaneous tests.

       Additionally, any calls to `flushall` or any other cross-database calls **will** still
       represent cross-test state.

       Finally, the above notes are purely describing the current implementation, and should not
       be assumed. In the future, the current database selection mechanism may change, or
       databases may not be used altogether.

    Args:
        scope (str): The scope of the fixture can be specified by the user, defaults to "function".
        decode_responses (bool): Whether to decode the responses from redis.

    Raises:
        KeyError: If any additional arguments are provided to the function than what is necessary.
    """

    @pytest.fixture(scope=scope)
    def _(request, pmr_redis_container, pmr_redis_config):
        database_number = 0
        if hasattr(request.config, "workerinput"):
            worker_input = request.config.workerinput
            worker_id = worker_input["workerid"]  # For example "gw0".
            database_number = int(worker_id[2:])

        if database_number >= 16:
            raise ValueError(
                "The redis fixture currently only supports up to 16 parallel executions"
            )

        db = redis.Redis(
            host=pmr_redis_config.host,
            port=pmr_redis_config.port,
            db=database_number,
            decode_responses=decode_responses or pmr_redis_config.decode_responses,
        )
        db.flushdb()

        Credentials.assign_from_credentials(
            db,
            drivername="redis",
            host=pmr_redis_config.host,
            port=pmr_redis_config.port,
            database=database_number,
            username=None,
            password=None,
        )
        return db

    return _



from __future__ import annotations

import io
from dataclasses import dataclass
from typing import BinaryIO, ClassVar, Iterable, TextIO, TYPE_CHECKING, Union

from pytest_mock_resources.action import AbstractAction, validate_actions

if TYPE_CHECKING:
    from pytest_mock_resources.fixture.moto.base import Session


ObjectContent = Union[bytes, str, BinaryIO, TextIO]


@dataclass
class S3Bucket(AbstractAction):
    fixtures: ClassVar[tuple[str, ...]] = ("moto",)
    static_safe: ClassVar[bool] = True

    name: str

    def object(self, key: str, data: ObjectContent):
        return S3Object(self, key, data)

    def apply(self, session: Session):
        client = session.resource("s3")
        client.create_bucket(Bucket=self.name)


@dataclass
class S3Object(AbstractAction):
    fixtures: ClassVar[tuple[str, ...]] = ("moto",)
    static_safe: ClassVar[bool] = True

    bucket: str | S3Bucket
    key: str
    data: ObjectContent
    encoding: str = "utf-8"

    def apply(self, session: Session):
        resource = session.resource("s3")
        bucket_name = self.bucket.name if isinstance(self.bucket, S3Bucket) else self.bucket

        if isinstance(self.data, str):
            data = io.BytesIO(self.data.encode(self.encoding))
        elif isinstance(self.data, bytes):
            data = io.BytesIO(self.data)
        elif isinstance(self.data, io.StringIO):
            data = io.BytesIO(self.data.getvalue().encode(self.encoding))
        elif isinstance(self.data, io.BytesIO):
            data = io.BytesIO(self.data.getbuffer().tobytes())
        else:
            raise NotImplementedError()

        resource.Object(bucket_name=bucket_name, key=self.key).upload_fileobj(data)


MotoAction = Union[S3Bucket, S3Object]


def apply_ordered_actions(session: Session, ordered_actions: Iterable[MotoAction]):
    validate_actions(ordered_actions, fixture="moto")
    for action in ordered_actions:
        action.apply(session)



from typing import ClassVar, Iterable

from pytest_mock_resources.container.postgres import PostgresConfig


class RedshiftConfig(PostgresConfig):
    """Define the configuration object for Redshift.

    Args:
        image (str): The docker image:tag specifier to use for Redshift containers.
            Defaults to :code:`"postgres:9.6.10-alpine"`.
        host (str): The hostname under which a mounted port will be available.
            Defaults to :code:`"localhost"`.
        port (int): The port to bind the container to.
            Defaults to :code:`5532`.
        ci_port (int): The port to bind the container to when a CI environment is detected.
            Defaults to :code:`5432`.
        username (str): The username of the root Redshift user
            Defaults to :code:`"user"`.
        password (str): The password of the root Redshift password
            Defaults to :code:`"password"`.
        root_database (str): The name of the root Redshift database to create.
            Defaults to :code:`"dev"`.
        drivername (str): The sqlalchemy driver to use
            Defaults to :code:`"postgresql+psycopg2"`.
    """

    name = "redshift"
    _fields: ClassVar[Iterable] = {
        "image",
        "host",
        "port",
        "ci_port",
        "username",
        "password",
        "root_database",
        "drivername",
    }
    _fields_defaults: ClassVar[dict] = {
        "image": "postgres:9.6.10-alpine",
        "port": 5532,
        "ci_port": 5432,
        "username": "user",
        "password": "password",
        "root_database": "dev",
        "drivername": None,
    }



from pytest_mock_resources.fixture.mongo import (
    create_mongo_fixture,
    pmr_mongo_config,
    pmr_mongo_container,
)
from pytest_mock_resources.fixture.moto import (
    create_moto_fixture,
    pmr_moto_config,
    pmr_moto_container,
    S3Bucket,
    S3Object,
)
from pytest_mock_resources.fixture.mysql import (
    create_mysql_fixture,
    pmr_mysql_config,
    pmr_mysql_container,
)
from pytest_mock_resources.fixture.postgresql import (
    create_postgres_fixture,
    pmr_postgres_config,
    pmr_postgres_container,
)
from pytest_mock_resources.fixture.redis import (
    create_redis_fixture,
    pmr_redis_config,
    pmr_redis_container,
)
from pytest_mock_resources.fixture.redshift import (
    create_redshift_fixture,
    pmr_redshift_config,
    pmr_redshift_container,
)
from pytest_mock_resources.fixture.sqlite import create_sqlite_fixture

__all__ = [
    "S3Bucket",
    "S3Object",
    "create_mongo_fixture",
    "create_moto_fixture",
    "create_mysql_fixture",
    "create_postgres_fixture",
    "create_redis_fixture",
    "create_redshift_fixture",
    "create_sqlite_fixture",
    "pmr_mongo_config",
    "pmr_mongo_container",
    "pmr_moto_config",
    "pmr_moto_container",
    "pmr_mysql_config",
    "pmr_mysql_container",
    "pmr_postgres_config",
    "pmr_postgres_container",
    "pmr_redis_config",
    "pmr_redis_container",
    "pmr_redshift_config",
    "pmr_redshift_container",
]



import pytest

from pytest_mock_resources.container.base import get_container
from pytest_mock_resources.container.redshift import RedshiftConfig
from pytest_mock_resources.fixture.base import asyncio_fixture, generate_fixture_id
from pytest_mock_resources.fixture.postgresql import _async_fixture, _sync_fixture
from pytest_mock_resources.patch.redshift import psycopg2, sqlalchemy


@pytest.fixture(scope="session")
def pmr_redshift_config():
    """Override this fixture with a :class:`RedshiftConfig` instance to specify different defaults.

    Note that, by default, redshift uses a postgres container.

    Examples:
        >>> @pytest.fixture(scope='session')
        ... def pmr_redshift_config():
        ...     return RedshiftConfig(image="postgres:9.6.10", root_database="foo")
    """
    return RedshiftConfig()


@pytest.fixture(scope="session")
def pmr_redshift_container(pytestconfig, pmr_redshift_config):
    yield from get_container(pytestconfig, pmr_redshift_config)


def create_redshift_fixture(
    *ordered_actions,
    scope="function",
    tables=None,
    session=None,
    async_=False,
    createdb_template="template1",
    engine_kwargs=None,
    template_database=True,
    actions_share_transaction=None,
):
    """Produce a Redshift fixture.

    Any number of fixture functions can be created. Under the hood they will all share the same
    database server.

    Note that, by default, redshift uses a postgres container as the database server
    and attempts to reintroduce approximations of Redshift features, such as
    S3 COPY/UNLOAD, redshift-specific functions, and other specific behaviors.

    Arguments:
        ordered_actions: Any number of ordered actions to be run on test setup.
        scope: Passthrough pytest's fixture scope.
        tables: Subsets the tables created by `ordered_actions`. This is generally
            most useful when a model-base was specified in `ordered_actions`.
        session: Whether to return a session instead of an engine directly. This can
            either be a bool or a callable capable of producing a session.
        async_: Whether to return an async fixture/client.
        createdb_template: The template database used to create sub-databases. "template1" is the
            default chosen when no template is specified.
        engine_kwargs: Optional set of kwargs to send into the engine on creation.
        template_database: Defaults to True. When True, amortizes the cost of performing database
            setup through `ordered_actions`, by performing them once into a postgres "template"
            database, then creating all subsequent per-test databases from that template.
        actions_share_transaction: When True, the transaction used by `ordered_actions` context
            will be the same as the one handed to the test function. This is required in order
            to support certain usages of `ordered_actions, such as the creation of temp tables
            through a `Statements` object. By default, this behavior is enabled for synchronous
            fixtures for backwards compatibility; and disabled by default for
            asynchronous fixtures (the way v2-style/async features work in SQLAlchemy can lead
            to bad default behavior).
    """
    from pytest_mock_resources.fixture.redshift.udf import REDSHIFT_UDFS

    fixture_id = generate_fixture_id(enabled=template_database, name="pg")

    ordered_actions = (*ordered_actions, REDSHIFT_UDFS)
    engine_kwargs_ = engine_kwargs or {}

    engine_manager_kwargs = {
        "ordered_actions": ordered_actions,
        "tables": tables,
        "createdb_template": createdb_template,
        "session": session,
        "fixture_id": fixture_id,
        "actions_share_transaction": actions_share_transaction,
    }

    @pytest.fixture(scope=scope)
    def _sync(*_, pmr_redshift_container, pmr_redshift_config):
        for engine, conn in _sync_fixture(
            pmr_redshift_config,
            engine_manager_kwargs,
            engine_kwargs_,
            fixture="redshift",
        ):
            sqlalchemy.register_redshift_behavior(engine)
            with psycopg2.patch_connect(pmr_redshift_config, engine.url.database):
                yield conn

    async def _async(*_, pmr_redshift_container, pmr_redshift_config):
        fixture = _async_fixture(
            pmr_redshift_config,
            engine_manager_kwargs,
            engine_kwargs_,
            fixture="redshift",
        )
        async for engine, conn in fixture:
            sqlalchemy.register_redshift_behavior(engine.sync_engine)
            yield conn

    if async_:
        return asyncio_fixture(_async, scope=scope)
    return _sync



from typing import ClassVar, Iterable

from pytest_mock_resources.compat import redis
from pytest_mock_resources.config import DockerContainerConfig, fallback
from pytest_mock_resources.container.base import ContainerCheckFailed


class RedisConfig(DockerContainerConfig):
    """Define the configuration object for redis.

    Args:
        image (str): The docker image:tag specifier to use for redis containers.
            Defaults to :code:`"redis:5.0.7"`.
        host (str): The hostname under which a mounted port will be available.
            Defaults to :code:`"localhost"`.
        port (int): The port to bind the container to.
            Defaults to :code:`6380`.
        ci_port (int): The port to bind the container to when a CI environment is detected.
            Defaults to :code:`6379`.
        decode_responses (bool): Whether to decode responses from the server on the client.
            Defaults to :code:`False`.
    """

    name = "redis"

    _fields: ClassVar[Iterable] = {
        "image",
        "host",
        "port",
        "ci_port",
        "decode_responses",
    }
    _fields_defaults: ClassVar[dict] = {
        "image": "redis:5.0.7",
        "port": 6380,
        "ci_port": 6379,
        "decode_responses": False,
    }

    @fallback
    def decode_responses(self):
        raise NotImplementedError()

    def ports(self):
        return {6379: self.port}

    def check_fn(self):
        try:
            client = redis.Redis(host=self.host, port=self.port)
            client.ping()
        except redis.ConnectionError:
            raise ContainerCheckFailed(
                f"Unable to connect to a presumed Redis test container via given config: {self}"
            )



import fnmatch
import logging
from collections.abc import Callable
from dataclasses import dataclass, field
from typing import Dict, Iterable, List, Optional, Set, TypeVar, Union

import sqlalchemy
from sqlalchemy import MetaData, text
from sqlalchemy.orm import scoped_session, Session, sessionmaker
from sqlalchemy.sql.ddl import CreateSchema
from sqlalchemy.sql.schema import Table

from pytest_mock_resources import compat
from pytest_mock_resources.action import AbstractAction, validate_actions
from pytest_mock_resources.credentials import Credentials

log = logging.getLogger(__name__)


class Rows(AbstractAction):
    fixtures = ("mysql", "postgres", "redshift", "sqlite")
    static_safe = True

    def __init__(self, *rows):
        self.rows = rows

    def apply(self, conn):
        rows = self._get_stateless_rows(self.rows)

        if isinstance(conn, Session):
            session = conn
        else:
            session = Session(bind=conn)

        session.add_all(rows)
        commit(session)

    @staticmethod
    def _get_stateless_rows(rows):
        """Create rows that aren't associated with any other SQLAlchemy session."""
        stateless_rows = []
        for row in rows:
            row_args = row.__dict__
            row_args.pop("_sa_instance_state", None)

            stateless_row = type(row)(**row_args)

            stateless_rows.append(stateless_row)
        return stateless_rows


class Statements(AbstractAction):
    fixtures = ("mysql", "postgres", "redshift", "sqlite")
    static_safe = False

    def __init__(self, *statements):
        self.statements = statements

    def apply(self, conn):
        for statement in self.statements:
            if isinstance(statement, str):
                statement = text(statement)
            conn.execute(statement)


class StaticStatements(Statements):
    """A discriminator for statements which are safe to execute exactly once."""

    static_safe = True


StaticAction = Union[MetaData, compat.sqlalchemy.DeclarativeMeta, Rows, StaticStatements]
Action = Union[MetaData, compat.sqlalchemy.DeclarativeMeta, AbstractAction]
T = TypeVar("T", StaticAction, Action)


@dataclass
class EngineManager:
    dynamic_actions: Iterable[Action] = ()
    tables: Iterable = ()
    session: Union[bool, Session] = False
    static_actions: Iterable[StaticAction] = ()
    actions_share_transaction: Optional[bool] = None

    _ddl_created: Dict[MetaData, bool] = field(default_factory=dict)

    @classmethod
    def create(
        cls,
        dynamic_actions: Iterable[Action],
        *,
        fixture: Optional[str] = None,
        tables: Iterable = (),
        session: Union[bool, Session] = False,
        static_actions: Iterable[StaticAction] = (),
    ) -> "EngineManager":
        return cls(
            static_actions=normalize_actions(static_actions, fixture=fixture),
            dynamic_actions=normalize_actions(dynamic_actions, fixture=fixture),
            tables=tables,
            session=session,
        )

    def manage_sync(self, engine):
        try:
            if self.session:
                if isinstance(self.session, sessionmaker):
                    session_factory = self.session
                    session = session_factory(bind=engine)
                elif isinstance(self.session, Session):
                    session = self.session
                else:
                    session_factory = scoped_session(sessionmaker(bind=engine))
                    session = session_factory(bind=engine)

                try:
                    self.run_actions(session)
                    commit(session)

                    if self.actions_share_transaction is False:
                        engine.dispose()

                    Credentials.assign_from_connection(session)
                    yield engine, session
                finally:
                    session.close()
            else:
                with engine.begin() as conn:
                    self.run_actions(conn)
                    commit(conn)

                if self.actions_share_transaction is False:
                    engine.dispose()

                Credentials.assign_from_connection(engine)
                yield engine, engine

        finally:
            engine.dispose()

    async def manage_async(self, engine, session=None):
        from sqlalchemy.ext.asyncio.session import AsyncSession

        try:
            if self.session:
                if isinstance(self.session, (sessionmaker, AsyncSession)):
                    session_factory = self.session
                else:
                    session_factory = sessionmaker(
                        expire_on_commit=False,
                        class_=compat.sqlalchemy.asyncio.AsyncSession,
                    )

                async with session_factory(bind=engine) as session:
                    await session.run_sync(self.run_actions)
                    if not self.actions_share_transaction:
                        await session.commit()
                        await session.close()

                    Credentials.assign_from_connection(engine.sync_engine)
                    yield engine, session
            else:
                async with engine.begin() as conn:
                    await conn.run_sync(self.run_actions)
                    await conn.commit()

                if not self.actions_share_transaction:
                    await engine.dispose()

                Credentials.assign_from_connection(engine.sync_engine)
                yield engine, engine
        finally:
            await engine.dispose()

    def run_actions(self, conn):
        self.run_static_actions(conn)
        self.run_dynamic_actions(conn)

    def run_static_actions(self, conn):
        for action in self.static_actions:
            self.execute_action(conn, action, allow_function=False)

    def run_dynamic_actions(self, conn):
        for action in self.dynamic_actions:
            self.execute_action(conn, action, allow_function=True)

    def _create_schemas(self, conn, metadata):
        if metadata in self._ddl_created:
            return

        all_schemas = {table.schema for table in metadata.tables.values() if table.schema}

        for schema in all_schemas:
            if hasattr(conn, "dialect"):
                dialect = conn.dialect
            else:
                dialect = conn.connection().dialect

            if dialect.default_schema_name == schema:
                continue

            statement = CreateSchema(schema)
            conn.execute(statement)

    def _create_tables(self, conn, metadata):
        if isinstance(conn, Session):
            conn = conn.connection()

        if not self.tables:
            metadata.create_all(conn)
            return

        table_objects = {
            table_object
            for table in self.tables
            for table_object in identify_matching_tables(metadata, table)
        }

        metadata.create_all(conn, tables=list(table_objects))

    def create_ddl(self, conn, metadata):
        self._create_schemas(conn, metadata)
        self._create_tables(conn, metadata)
        self._ddl_created[metadata] = True

    def execute_action(self, conn, action, allow_function=False):
        if isinstance(action, MetaData):
            self.create_ddl(conn, action)
        elif isinstance(action, AbstractAction):
            action.apply(conn)
        elif allow_function and callable(action):
            action(conn)
            commit(conn)


def normalize_actions(
    ordered_actions: Iterable[T], *, fixture: Optional[str] = None
) -> Iterable[T]:
    validate_actions(
        ordered_actions,
        fixture=fixture,
        additional_types=(compat.sqlalchemy.DeclarativeMeta, Callable, MetaData),
    )

    # Keep track of metadata we've seen to ensure it's only added once per
    # instance, regardless of `Row` references.
    unique_metadata: Set[MetaData] = set()
    normalized_actions: List[T] = []
    for action in ordered_actions:
        metadata = compat.sqlalchemy.extract_model_base_metadata(action)
        if metadata:
            normalized_actions.append(metadata)

        elif isinstance(action, Rows):
            new_metadata = {row.metadata for row in action.rows} - unique_metadata
            unique_metadata |= new_metadata
            normalized_actions.extend(list(new_metadata))
            normalized_actions.append(action)

        elif isinstance(action, MetaData):
            if action not in unique_metadata:
                normalized_actions.append(action)
        elif isinstance(action, AbstractAction) or callable(action):
            normalized_actions.append(action)

    return normalized_actions


def bifurcate_actions(ordered_actions):
    static_actions = []
    dynamic_actions = []

    static = True
    for action in ordered_actions:
        static_safe = isinstance(action, MetaData) or (
            isinstance(action, AbstractAction) and action.static_safe
        )

        if static:
            if static_safe:
                static_actions.append(action)
            else:
                static = False
                dynamic_actions.append(action)
        else:
            dynamic_actions.append(action)

    return (static_actions, dynamic_actions)


def identify_matching_tables(metadata, table_specifier):
    if isinstance(table_specifier, compat.sqlalchemy.DeclarativeMeta):
        return [table_specifier.__table__]

    if isinstance(table_specifier, Table):
        return [table_specifier]

    tables = [
        table
        for table_name, table in metadata.tables.items()
        if fnmatch.fnmatch(table_name, table_specifier)
    ]

    if tables:
        return tables

    table_names = ", ".join(sorted(metadata.tables.keys()))
    raise ValueError(
        f'Could not identify any tables matching "{table_specifier}" from: {table_names}'
    )


def commit(conn):
    try:
        if isinstance(conn, Session):
            return conn.commit()

        return conn.execute(text("COMMIT"))
    except sqlalchemy.exc.InvalidRequestError:
        # In autocommit mode, we won't be able to commit.
        pass



from typing import Union

from sqlalchemy import event
from sqlalchemy.sql.base import Executable

from pytest_mock_resources.compat import sqlparse
from pytest_mock_resources.patch.redshift.mock_s3_copy import mock_s3_copy_command
from pytest_mock_resources.patch.redshift.mock_s3_unload import mock_s3_unload_command


def register_redshift_behavior(engine):
    """Substitute the default execute method with a custom execute for copy and unload command."""
    event.listen(engine, "before_execute", receive_before_execute, retval=True)
    event.listen(engine, "before_cursor_execute", receive_before_cursor_execute, retval=True)


def receive_before_execute(
    conn, clauseelement: Union[Executable, str], multiparams, params, execution_options=None
):
    """Handle the `before_execute` event.

    Specifically, this only needs to handle the parsing of multiple statements into
    individual cursor executions. Only the final statement's return value will be
    returned.
    """
    if isinstance(clauseelement, Executable):
        return clauseelement, multiparams, params

    *statements, final_statement = parse_multiple_statements(clauseelement)

    cursor = conn.connection.cursor()
    for statement in statements:
        cursor.execute(statement, *multiparams, **params)

    return final_statement, multiparams, params


def receive_before_cursor_execute(_, cursor, statement: str, parameters, context, executemany):
    """Handle the `before_cursor_execute` event.

    This is where we add support for custom features such as redshift COPY/UNLOAD because
    the query has already been rendered (into a string) at this point.

    Notably, COPY/UNLOAD need to perform extra non-sql behavior and potentially execute
    more than a single query and the interface requires that we return a query. Thus,
    we return a no-op query to be executed by sqlalchemy for certain kinds of supported
    extra features.
    """
    normalized_statement = _preprocess(statement).lower()
    if normalized_statement.startswith("unload"):
        mock_s3_unload_command(statement, cursor)
        return "SELECT 1", {}

    if normalized_statement.startswith("copy"):
        mock_s3_copy_command(statement, cursor)
        context.should_autocommit = True
        return "SELECT 1", {}
    return statement, parameters


def parse_multiple_statements(statement: str):
    """Split the given sql statement into a list of individual sql statements."""
    processed_statement = _preprocess(statement)
    return [str(statement) for statement in sqlparse.split(processed_statement)]


def _preprocess(statement: str):
    """Preprocess the input statement."""
    statement = statement.strip()
    # Replace any occourance of " with '.
    statement = statement.replace('"', "'")
    if statement[-1] != ";":
        statement += ";"
    return statement



from pytest_mock_resources.container import (
    get_container,
    MongoConfig,
    MotoConfig,
    MysqlConfig,
    PostgresConfig,
    RedisConfig,
    RedshiftConfig,
)
from pytest_mock_resources.credentials import Credentials
from pytest_mock_resources.fixture import (
    create_mongo_fixture,
    create_moto_fixture,
    create_mysql_fixture,
    create_postgres_fixture,
    create_redis_fixture,
    create_redshift_fixture,
    create_sqlite_fixture,
    pmr_mongo_config,
    pmr_mongo_container,
    pmr_moto_config,
    pmr_moto_container,
    pmr_mysql_config,
    pmr_mysql_container,
    pmr_postgres_config,
    pmr_postgres_container,
    pmr_redis_config,
    pmr_redis_container,
    pmr_redshift_config,
    pmr_redshift_container,
    S3Bucket,
    S3Object,
)
from pytest_mock_resources.hooks import (  # noqa
    pytest_addoption,
    pytest_configure,
    pytest_itemcollected,
    pytest_sessionfinish,
)
from pytest_mock_resources.sqlalchemy import Rows, Statements, StaticStatements

__all__ = [
    "Credentials",
    "MongoConfig",
    "MotoConfig",
    "MysqlConfig",
    "PostgresConfig",
    "RedisConfig",
    "RedshiftConfig",
    "Rows",
    "Statements",
    "StaticStatements",
    "S3Bucket",
    "S3Object",
    "create_mongo_fixture",
    "create_moto_fixture",
    "create_mysql_fixture",
    "create_postgres_fixture",
    "create_redis_fixture",
    "create_redshift_fixture",
    "create_sqlite_fixture",
    "get_container",
    "pmr_mongo_config",
    "pmr_mongo_container",
    "pmr_moto_config",
    "pmr_moto_container",
    "pmr_mysql_config",
    "pmr_mysql_container",
    "pmr_postgres_config",
    "pmr_postgres_container",
    "pmr_redis_config",
    "pmr_redis_container",
    "pmr_redshift_config",
    "pmr_redshift_container",
]



"""Adapt SQLite to act more in line with typical postgresql usage.

SQLite is a desirable test database target because it introduces much less latency across
tests, is parallelizable, and has a lower minimum fixed test-startup cost.

Unfortunately, SQLite and postgresql do not behave identically in all cases. However, in
many (typically CRUD) API type database designs, there is no need for postgresql-specific
features, and SQLite can act as an appropriate stand in.

Where possible though, this module changes the default SQLite behavior to more closely
mimic postgresql, so as to increase the number of places where SQLite can frictionlessly
stand in for postgresql.
"""
import contextlib
import datetime
import json
import warnings

import pytest
from sqlalchemy import create_engine, dialects, event
from sqlalchemy.dialects.postgresql import JSON, JSONB
from sqlalchemy.dialects.sqlite import base as sqlite_base
from sqlalchemy.dialects.sqlite.pysqlite import SQLiteDialect_pysqlite
from sqlalchemy.exc import SAWarning
from sqlalchemy.ext.compiler import compiles
from sqlalchemy.sql import sqltypes

from pytest_mock_resources.sqlalchemy import EngineManager


class PMRSQLiteDDLCompiler(sqlite_base.SQLiteDDLCompiler):
    """Add features sqlite doesn't support by default.

    Adds:
     * Handling `CREATE SCHEMA`.
    """

    def visit_create_schema(self, create):
        schema = self.preparer.format_schema(create.element)
        return "ATTACH DATABASE ':memory:' AS " + schema

    def visit_drop_schema(self, create):
        schema = self.preparer.format_schema(create.element)
        return "DETACH DATABASE " + schema


class UTC(datetime.tzinfo):
    """UTC timezone.

    Blame python2:
    https://docs.python.org/2/library/datetime.html#datetime.tzinfo.fromutc
    """

    ZERO = datetime.timedelta(0)

    def utcoffset(self, dt):
        return self.ZERO

    def dst(self, dt):
        return self.ZERO

    def tzname(self, dt):
        return "UTC"


utc = UTC()


class PMRDateTime(sqlite_base.DATETIME):
    def bind_processor(self, dialect):
        processor = super().bind_processor(dialect)

        def process(value):
            if isinstance(value, datetime.datetime):
                if value.tzinfo is None:
                    value = value.replace(tzinfo=utc)
                result = value.astimezone(utc)
                return processor(result)
            return None

        return process

    def result_processor(self, dialect, coltype):
        processor = super().result_processor(dialect, coltype)

        def process(value):
            result = processor(value)
            if self.timezone:
                if result is not None and result.tzinfo is None:
                    result = result.replace(tzinfo=utc)
            return result

        return process


PostgresLikeSQLitePDialect = None


def make_postgres_like_sqlite_dialect():
    if not hasattr(sqlite_base, "JSON"):
        raise RuntimeError(
            "Must have sqlalchemy>=1.3.0 in order to make use of the postgres-like sqlite dialect."
        )

    class PostgresLikeTypeCompiler(sqlite_base.SQLiteTypeCompiler):
        """Add features sqlite doesn't support by default.

        Adds:
         * Handling of postgres column types: JSON, JSONB
        """

        @compiles(JSON, "pmrsqlite")
        @compiles(JSONB, "pmrsqlite")
        @compiles(sqltypes.JSON, "pmrsqlite")
        def compile_json(type_, compiler, **kwargs):  # noqa: N805
            return "BLOB"

    class PostgresLikeDialect(SQLiteDialect_pysqlite):
        """Define the dialect that collects all postgres-like adaptations."""

        name = "pmrsqlite"

        supports_statement_cache = True
        ddl_compiler = PMRSQLiteDDLCompiler
        type_compiler = PostgresLikeTypeCompiler
        colspecs: dict = {  # noqa: RUF012
            sqltypes.Date: sqlite_base.DATE,
            sqltypes.DateTime: PMRDateTime,
            sqltypes.JSON: sqlite_base.JSON,
            sqltypes.JSON.JSONIndexType: sqlite_base.JSONIndexType,
            sqltypes.JSON.JSONPathType: sqlite_base.JSONPathType,
            sqltypes.Time: sqlite_base.TIME,
        }

        def _json_serializer(self, value):
            return json.dumps(value, sort_keys=True)

    global PostgresLikeSQLitePDialect
    PostgresLikeSQLitePDialect = PostgresLikeDialect
    return PostgresLikeSQLitePDialect


def enable_foreign_key_checks(dbapi_connection, connection_record):
    """Enable foreign key constraint checks.

    SQLite supports FOREIGN KEY syntax when emitting CREATE statements for tables,
    however by default these constraints have no effect on the operation of the table.
    https://docs.sqlalchemy.org/en/latest/dialects/sqlite.html#foreign-key-support
    """
    cursor = dbapi_connection.cursor()
    cursor.execute("PRAGMA foreign_keys=ON")
    cursor.close()


def do_connect(dbapi_connection, connection_record):
    """Disable pysqlite's emitting of the BEGIN statement entirely.

    Also stops it from emitting COMMIT before any DDL.

    https://docs.sqlalchemy.org/en/20/dialects/sqlite.html#pysqlite-serializable
    """
    dbapi_connection.isolation_level = None


def do_begin(conn):
    """Emit our own BEGIN.

    SQLite lazily emits begin on first write operation by default. This impacts our ability
    to correctly manage the session in `EngineManager`.

    https://docs.sqlalchemy.org/en/20/dialects/sqlite.html#pysqlite-serializable
    """
    try:
        conn.exec_driver_sql("BEGIN")
    except AttributeError:  # pragma: no cover
        conn.execute("BEGIN")


@contextlib.contextmanager
def filter_sqlalchemy_warnings(decimal_warnings_enabled=True):
    with warnings.catch_warnings():
        if decimal_warnings_enabled:
            warnings.filterwarnings(
                "ignore",
                (
                    r"^Dialect pmrsqlite\+pysqlite does \*not\* support Decimal objects natively\, and "
                    r"SQLAlchemy must convert from floating point - rounding errors and other issues may occur\. "
                    r"Please consider storing Decimal numbers as strings or integers on this platform for "
                    r"lossless storage\.$"
                ),
                SAWarning,
            )

        yield


def _database_producer():
    i = 1
    while True:
        yield f"db{i}"
        i += 1


_database_names = _database_producer()


def create_sqlite_fixture(
    *ordered_actions,
    scope="function",
    tables=None,
    session=None,
    decimal_warnings=False,
    postgres_like=True,
):
    """Produce a SQLite fixture.

    Any number of fixture functions can be created. Under the hood they will all share the same
    database server.

    Arguments:
        ordered_actions: Any number of ordered actions to be run on test setup.
        scope: Passthrough pytest's fixture scope.
        tables: Subsets the tables created by `ordered_actions`. This is generally
            most useful when a model-base was specified in `ordered_actions`.
        session: Whether to return a session instead of an engine directly. This can
            either be a bool or a callable capable of producing a session.
        decimal_warnings: Whether to show sqlalchemy decimal warnings related to precision loss. The
            default `False` suppresses these warnings.
        postgres_like: Whether to add extra SQLite features which attempt to mimic postgres
            enough to stand in for it for testing.
    """
    dialect_name = "sqlite"

    if postgres_like:
        dialect = make_postgres_like_sqlite_dialect()
        dialect_name = dialect.name
        registry = getattr(dialects, "registry")
        registry.register(f"sqlite.{dialect_name}", __name__, "PostgresLikeSQLitePDialect")

    driver_name = f"sqlite+{dialect_name}"

    @pytest.fixture(scope=scope)
    def _():
        # XXX: Ideally we eventually make use of the shared memory cache to enable connecting by
        #      credentials with sqlite.
        # database_name = "file:{}?mode=memory&cache=shared".format(next(_database_names))
        database_name = ""

        raw_engine = create_engine(f"{driver_name}:///{database_name}")

        # This *must* happen before the connection occurs (implicitly in `EngineManager`).
        event.listen(raw_engine, "connect", enable_foreign_key_checks)

        # https://docs.sqlalchemy.org/en/20/dialects/sqlite.html#pysqlite-serializable
        event.listen(raw_engine, "connect", do_connect)
        event.listen(raw_engine, "begin", do_begin)

        engine_manager = EngineManager.create(
            fixture="sqlite",
            dynamic_actions=ordered_actions,
            tables=tables,
            session=session,
        )
        for _, conn in engine_manager.manage_sync(raw_engine):
            with filter_sqlalchemy_warnings(decimal_warnings_enabled=(not decimal_warnings)):
                yield conn

        event.remove(raw_engine, "connect", enable_foreign_key_checks)

    return _



from typing import ClassVar, Iterable

from pytest_mock_resources.config import DockerContainerConfig
from pytest_mock_resources.container.base import ContainerCheckFailed


class MotoConfig(DockerContainerConfig):
    """Define the configuration object for moto.

    Args:
        image (str): The docker image:tag specifier to use for postgres containers.
            Defaults to :code:`"postgres:9.6.10-alpine"`.
        host (str): The hostname under which a mounted port will be available.
            Defaults to :code:`"localhost"`.
        port (int): The port to bind the container to.
            Defaults to :code:`5532`.
        ci_port (int): The port to bind the container to when a CI environment is detected.
            Defaults to :code:`5432`.
    """

    name = "moto"

    _fields: ClassVar[Iterable] = {"image", "host", "port"}
    _fields_defaults: ClassVar[dict] = {
        "image": "motoserver/moto:4.0.6",
        "port": 5555,
    }

    def ports(self):
        return {5000: self.port}

    def check_fn(self):
        import requests

        try:
            url = endpoint_url(self)
            requests.get(url, timeout=60)
        except requests.exceptions.RequestException:
            raise ContainerCheckFailed(
                f"Unable to connect to a presumed moto test container via given config: {self}"
            )


def endpoint_url(config):
    return f"http://{config.host}:{config.port}"
